"""
This module provides functions to read in an XES formatted event log in various forms.
"""

from dataclasses import dataclass
from enum import Enum
from os import path
from datetime import datetime
from typing import List, Mapping
from copy import deepcopy

from xml.etree.ElementTree import Element, parse

from pmkoalas.simple import EventLog, Trace
from pmkoalas.complex import ComplexEvent, ComplexTrace, ComplexEventLog
from pmkoalas._logging import debug, info, enable_logging
from pmkoalas.xes import XES_CONCEPT,XES_TIME,XES_XML_NAMESPACE

from pmkoalas.xes_export import XES_STRING_TAG, XES_TIME_TAG , XES_INT_TAG
from pmkoalas.xes_export import XES_FLOAT_TAG, XES_BOOLEAN_TAG

class XesType(Enum):
    STRING = XES_STRING_TAG
    DATE = XES_TIME_TAG
    INT = XES_INT_TAG
    FLOAT = XES_FLOAT_TAG
    BOOLEAN = XES_BOOLEAN_TAG
    UNKNOWN = "NaN"


def find_xes_type(tag:str) -> XesType:
    if XesType.STRING.value in tag:
        return XesType.STRING
    elif XesType.DATE.value in tag:
        return XesType.DATE
    elif XesType.INT.value in tag: 
        return XesType.INT
    elif XesType.FLOAT.value in tag:
        return XesType.FLOAT
    elif XesType.BOOLEAN.value in tag:
        return XesType.BOOLEAN

    return XesType.UNKNOWN

@dataclass
class XesAttribute:
    type:XesType 
    key:str
    value:str

    def get(self) -> object:
        if self.type == XesType.STRING:
            return self.value.__str__()
        elif self.type == XesType.DATE:
            return datetime.fromisoformat(self.value)
        elif self.type == XesType.INT:
            return int(self.value.__str__())
        elif self.type == XesType.FLOAT:
            pass_one = float(self.value.__str__())
            pass_two = f"{pass_one:.8f}"
            return float(pass_two)
        elif self.type == XesType.BOOLEAN:
            val = str(self.value).lower()
            if (val == 'true' or val == '1'):
                return True
            elif (val == 'false' or val == '0'):
                return False
        return self.value.__str__()

@dataclass
class EventExtract:
    event_order:int 
    label:XesAttribute 
    sorter:XesAttribute
    map:Mapping[str,XesAttribute] = None

    def get_label(self) -> str:
        return self.label.get()
    
    def get_data(self) -> Mapping[str,XesAttribute]:
        if self.map == None:
            return dict()
        else:
            return deepcopy(self.map)

    def get_sorter(self) -> object:
        return self.sorter.get()

    def __str__(self) -> str:
        return f"{self.event_order}-{self.get_label()}-{self.get_sorter()}"
    
def find_element(root:Element, find:str, use_namespace:bool) -> List[Element]:
    if (use_namespace):
        return [ event for event in root.findall(f"xes:{find}", 
                        XES_XML_NAMESPACE)]
    else: 
        return [ event for event in root.findall(f"{find}")]

@enable_logging
def read_xes_complex(filepath:str,
                    label_attribute=XES_CONCEPT) -> ComplexEventLog:
    """
    Reads an XES formatted event log and creates a simplified event log
    object. Traces from the event log are sorted by the sort_attribute
    (time:timestamp by default) before making the sequence of labels
    (concept:name by default).

    Parameters
    ----------
    filepath: `str`
    \t the filepath to the xes file to read.
    sort_attribute: `str`=`time:timestamp`
    \t the xes attribute to sort on.
    label_attribute: `str`=`concept:name`
    \t the xes attribute for the process label for an event
    debug: `bool`=`True`
    \t whether to print debug messages or not
    sort: `bool`=`False`
    \t whether to sort activity labels by another xes attribute or not
    """ 

    # check that file exists
    if not path.exists(filepath):
        raise FileNotFoundError("event log file not found at :: "+filepath)
    
    # parse traces
    xml_tree = parse(filepath)
    log = xml_tree.getroot()

    if (log == None):
        raise ValueError("Unable to find log element in xml structure")
    
    # find name of log
    log_attrs = log.find(".*[@key='concept:name']")

    name = "Unknown Event log"
    if log_attrs != None:
        name = log_attrs.attrib.get("value")
        debug(f"extracted event log name :: {name}")

    # check with namespace
    ns_traces = [ trace for trace in log.findall("xes:trace",
     XES_XML_NAMESPACE)]
    #  check without namespace
    nns_traces = [ trace for trace in log.findall("trace",
     XES_XML_NAMESPACE)]
    #  decide what to do with namespace
    use_namespace = True
    if (len(ns_traces) == nns_traces):
        traces = ns_traces 
        del nns_traces
    elif (len(ns_traces) < len(nns_traces)):
        traces = nns_traces
        use_namespace = False
        del ns_traces 
    else: 
        traces = ns_traces
        del ns_traces

    info(f"parsing {len(traces)} traces ...")
    # extract the following from a trace,
    # a sequence of activity labels
    # sort traces by time:timestamp before 
    extracted_traces = []
    for trace in traces:
        trace_ins = [] # each element is a EventExtract
        events = find_element(trace, "event", use_namespace)
        for id,event in enumerate(events):
            label = None 
            sorter = None 
            map = dict()
            for child in event.iter():
                key = child.attrib.get('key')
                # print(key)
                if (key == label_attribute):
                    label = XesAttribute(find_xes_type(child.tag), 
                                        key, child.attrib.get("value"))
                else:
                    if (key == None):
                        continue
                    map[key] = XesAttribute(
                        find_xes_type(child.tag),
                        key, 
                        child.attrib.get("value")                        
                    ).get()
            extract = EventExtract(id, label , sorter, map)
            trace_ins.append(extract)
        # handle complex event construction
        # collect any trace level attributes
        trace_map = dict()
        for child in trace:
            key = child.attrib.get('key')
            if (key == None):
                continue
            trace_map[key] = XesAttribute(
                        find_xes_type(child.tag),
                        key, 
                        child.attrib.get("value")                        
                    ).get()
        # build complex events and trace
        trace_ins = ComplexTrace(
            [ ComplexEvent(t.get_label(),t.get_data()) for t in trace_ins ],
             data=trace_map)
        # store complex trace
        extracted_traces.append(trace_ins) 
    # build log-level attributes
    log_map = dict()
    for child in log:
        key = child.attrib.get('key')
        if (key == None):
            continue
        log_map[key] = XesAttribute(
                    find_xes_type(child.tag),
                    key, 
                    child.attrib.get("value")                        
                ).get()
    return ComplexEventLog(extracted_traces, name=name, data=log_map)

@enable_logging
def read_xes_simple(filepath:str, label_attribute=XES_CONCEPT) -> EventLog:
    """
    Reads an XES formatted event log and creates a simplified event log 
    object. Traces from the event log are kept in document order before 
    making the sequence of labels (concept:name by default).

    Parameters
    ----------
    filepath: `str`
    \t the filepath to the xes file to read.
    label_attribute: `str`=`concept:name`
    \t the xes attribute for the process label for an event
    """

    # check that file exists
    if not path.exists(filepath):
        raise FileNotFoundError("event log file not found at :: "+filepath)

    # parse traces
    xml_tree = parse(filepath)
    log = xml_tree.getroot()

    if (log == None):
        raise ValueError("Unable to find log element in xml structure")

    # find name of log
    log_attrs = log.find(".*[@key='concept:name']")

    name = "Unknown Event log"
    if log_attrs != None:
        name = log_attrs.attrib.get("value")
        debug(f"extracted event log name :: {name}")

    # check with namespace
    ns_traces = [ trace for trace in log.findall("xes:trace",
     XES_XML_NAMESPACE)]
    #  check without namespace
    nns_traces = [ trace for trace in log.findall("trace",
     XES_XML_NAMESPACE)]
    #  decide what to do with namespace
    use_namespace = True
    if (len(ns_traces) == nns_traces):
        traces = ns_traces 
        del nns_traces
    elif (len(ns_traces) < len(nns_traces)):
        traces = nns_traces
        use_namespace = False
        del ns_traces 
    else: 
        traces = ns_traces
        del ns_traces

    info(f"parsing {len(traces)} traces ...")
    # extract the following from a trace,
    # a sequence of activity labels
    # sort traces by time:timestamp before 
    extracted_traces = []
    for trace in traces:
        trace_ins = [] # eache element is a EventExtract
        events = find_element(trace, "event", use_namespace)
        for id,event in enumerate(events):
            label = None 
            sorter = None 
            for child in event.iter():
                key = child.attrib.get('key')
                # print(key)
                if (key == label_attribute):
                    label = XesAttribute(find_xes_type(child.tag), 
                                        key, child.attrib.get("value"))
            extract = EventExtract(id, label , sorter)
            trace_ins.append(extract)
        trace_ins = Trace([ t.get_label() for t in trace_ins ])
        extracted_traces.append(trace_ins) 
    return EventLog(extracted_traces, name)




import unittest
from os import path 

from pmkoalas.read import read_xes_simple
from pmkoalas.simple import EventLog,Trace

SSMALL = path.join(".","tests","small_01.xes")
WSMALL = path.join(".","tests","small_02.xes")
OSMALL = path.join(".","tests","small_03.xes")

class DTLogTest(unittest.TestCase):

    def test_successful_read(self):
        try:
            log = read_xes_simple(SSMALL) 
        except:
            self.fail("Failed to parse log")

        real_log = EventLog( [
                Trace(["A","B","C","D","E"]),
                Trace(["A","B","C","D","E"]),
                Trace(["A","B","C","D","D"]),
                Trace(["A","A","A","B","E"]),
                Trace(["A","B","C","E","E"]),
            ])
        self.assertEqual(log, real_log)

    def test_unsucessful_read(self):
        try:
            log = read_xes_simple(WSMALL)
        except:
            return
        self.fail()

    def test_unsuccessful_label_change(self):
        try:
            log = read_xes_simple(SSMALL, label_attribute="foo:bah") 
        except:
            return 
        self.fail("Should fail when label attribute is not found!")

    def test_label_change(self):
        try:
            log = read_xes_simple(OSMALL, label_attribute="name") 
        except:
            self.fail("Failed to parse log after changing label attribute!")

        real_log = EventLog( [
                Trace(["E","D","C","B","A"]),
                Trace(["E","D","C","B","A"]),
                Trace(["D","D","C","B","A"]),
                Trace(["E","B","A","A","A"]),
                Trace(["E","E","C","B","A"]),
            ])
        self.assertEqual(log, real_log)

    def nottest_read_bad(self):
        with self.failUnlessRaises(Exception): 
            log = read_xes_simple(path.join(".","tests","small_bad.xes")) 
            print(f"{log}")

if __name__ == '__main__':
    unittest.main()




import unittest
from copy import deepcopy

from pmkoalas.discovery.alpha_miner import AlphaMinerInstance
from pmkoalas.discovery.alpha_miner import AlphaRelation,AlphaPair
from pmkoalas.discovery.alpha_miner import AlphaPlace,AlphaFlowRelation
from pmkoalas.discovery.alpha_miner import AlphaTransition
from pmkoalas.discovery.alpha_miner import AlphaSinkPlace,AlphaStartPlace
from pmkoalas.dtlog import convert

# simple log tests
LOG = convert( 
            "a b d",
            "a c d",
            "a b",
            "a c",
            "a b d e b d",
            "a c d e c d"
        )
START_ACTS = set(["a"])
END_ACTS = set(["d", "b", "c"])
ACTS = set(["a", "b", "c", "d", "e"])
MATRIX = { 
    "a" : {
        "a" : AlphaRelation("a","a",[]),
        "b" : AlphaRelation("a","b",[("a","b")]),
        "c" : AlphaRelation("a","c",[("a","c")]),
        "d" : AlphaRelation("a","d",[]),
        "e" : AlphaRelation("a","e",[])
    },
    "b" : {
        "a" : AlphaRelation("b","a",[("a","b")]),
        "b" : AlphaRelation("b","b",[]),
        "c" : AlphaRelation("b","c",[]),
        "d" : AlphaRelation("b","d",[("b","d")]),
        "e" : AlphaRelation("b","e",[("e","b")])
    },
    "c" : {
        "a" : AlphaRelation("c","a",[("a","c")]),
        "b" : AlphaRelation("c","b",[]),
        "c" : AlphaRelation("c","c",[]),
        "d" : AlphaRelation("c","d",[("c","d")]),
        "e" : AlphaRelation("c","e",[("e","c")])
    },
    "d" : {
        "a" : AlphaRelation("d","a",[]),
        "b" : AlphaRelation("d","b",[("b","d")]),
        "c" : AlphaRelation("d","c",[("c","d")]),
        "d" : AlphaRelation("d","d",[]),
        "e" : AlphaRelation("d","e",[("d","e")])
    },
    "e" : {
        "a" : AlphaRelation("e","a",[]),
        "b" : AlphaRelation("e","b",[("e","b")]),
        "c" : AlphaRelation("e","c",[("e","c")]),
        "d" : AlphaRelation("e","d",[("d","e")]),
        "e" : AlphaRelation("e","e",[])
    }
}

# following p173 and L5 in Process Mining (2016;2ND ED)
BOOK_LOG = convert( 
    "a b e f",
    "a b e f",
    "a b e c d b f",
    "a b e c d b f",
    "a b e c d b f",
    "a b c e d b f",
    "a b c e d b f",
    "a b c d e b f",
    "a b c d e b f",
    "a b c d e b f",
    "a b c d e b f",
    "a e b c d b f",
    "a e b c d b f",
    "a e b c d b f",
)
BOOK_TL = set(["a", "b", "c", "d", "e", "f"])
BOOK_TL_OUT = set([ AlphaTransition(t) for t in BOOK_TL])
BOOK_TI = set(["a"])
BOOK_TO = set(["f"])
BOOK_XL = set([ 
    AlphaPair(set(["a"]),set(["b"])),
    AlphaPair(set(["a"]), set(["e"])),
    AlphaPair(set(["b"]), set(["c"])),
    AlphaPair(set(["b"]), set(["f"])),
    AlphaPair(set(["c"]), set(["d"])),
    AlphaPair(set(["d"]), set(["b"])),
    AlphaPair(set(["e"]), set(["f"])),
    AlphaPair(set(["a","d"]), set(["b"])),
    AlphaPair(set(["b"]), set(["c","f"])),
])
BOOK_YL = [
    AlphaPair(set(["a"]), set(["e"])),
    AlphaPair(set(["c"]), set(["d"])),
    AlphaPair(set(["e"]), set(["f"])),
    AlphaPair(set(["a","d"]), set(["b"])),
    AlphaPair(set(["b"]), set(["c","f"])),
]
BOOK_PL = [ 
    AlphaPlace(BOOK_YL[0].__str__(), BOOK_YL[0].left, BOOK_YL[0].right),
    AlphaPlace(BOOK_YL[1].__str__(), BOOK_YL[1].left, BOOK_YL[1].right),
    AlphaPlace(BOOK_YL[2].__str__(), BOOK_YL[2].left, BOOK_YL[2].right),
    AlphaPlace(BOOK_YL[3].__str__(), BOOK_YL[3].left, BOOK_YL[3].right),
    AlphaPlace(BOOK_YL[4].__str__(), BOOK_YL[4].left, BOOK_YL[4].right),
    AlphaStartPlace(),
    AlphaSinkPlace(),
]
BOOK_FL = set([ 
    AlphaFlowRelation("1", AlphaTransition("a"), BOOK_PL[0]),
    AlphaFlowRelation("2", BOOK_PL[0], AlphaTransition("e")),
    AlphaFlowRelation("3", AlphaTransition("c"), BOOK_PL[1]),
    AlphaFlowRelation("4", BOOK_PL[1], AlphaTransition("d")),
    AlphaFlowRelation("5", AlphaTransition("e"), BOOK_PL[2]),
    AlphaFlowRelation("6", BOOK_PL[2], AlphaTransition("f")),
    AlphaFlowRelation("7", AlphaTransition("a"), BOOK_PL[3]),
    AlphaFlowRelation("8", AlphaTransition("d"), BOOK_PL[3]),
    AlphaFlowRelation("9", BOOK_PL[3], AlphaTransition("b")),
    AlphaFlowRelation("10", AlphaTransition("b"), BOOK_PL[4]),
    AlphaFlowRelation("11", BOOK_PL[4], AlphaTransition("c")),
    AlphaFlowRelation("12", BOOK_PL[4], AlphaTransition("f")),
    AlphaFlowRelation("13", BOOK_PL[5], AlphaTransition("a")),
    AlphaFlowRelation("14", AlphaTransition("f"), BOOK_PL[6]),    
])
BOOK_YL = set(BOOK_YL)
BOOK_PL = set(BOOK_PL)

class DTLogTest(unittest.TestCase):

    def test_creation(self):
        miner = AlphaMinerInstance(min_inst=5)
        miner = AlphaMinerInstance()
        miner = deepcopy(miner)

    def test_footprint_matrix(self):
        miner = AlphaMinerInstance()
        matrix = miner.mine_footprint_matrix(LOG)
        # test that footprint matrix has the right follows
        # test that footprint matrix has the right relations
        for col in ACTS:
            for row in ACTS:
                crelation = matrix[col][row]
                rrelation = MATRIX[col][row]
                # test one
                self.assertEqual(crelation._follows, 
                    rrelation._follows,
                    "computed and expected follows differ :: "+
                    f"{crelation._follows} vs {rrelation._follows}"
                )
                # test two
                self.assertEqual(crelation.relation(), 
                    rrelation.relation(),
                    "computed and expected alpha relation differ :: "+
                    f"{crelation} vs {rrelation}"
                )


    def test_step_one(self):
        # simple test
        miner = AlphaMinerInstance()
        out = miner._step_one(LOG)
        self.assertEqual(out, ACTS)

        # following p173 and L5 in Process Mining (2016;2ND ED)
        out = miner._step_one(BOOK_LOG)
        self.assertEqual(out, BOOK_TL)

    def test_step_two(self):
        # simple test
        miner = AlphaMinerInstance()
        out = miner._step_two(LOG)
        self.assertEqual(out, START_ACTS)  

        # following p173 and L5 in Process Mining (2016;2ND ED)
        out = miner._step_two(BOOK_LOG)
        self.assertEqual(out, BOOK_TI)

    def test_step_three(self):
        # simple test
        miner = AlphaMinerInstance()
        out = miner._step_three(LOG)
        self.assertEqual(out, END_ACTS) 

        # following p173 and L5 in Process Mining (2016;2ND ED)
        out = miner._step_three(BOOK_LOG)
        self.assertEqual(out, BOOK_TO)

    def test_step_four(self):
        miner = AlphaMinerInstance()
        # following p173 and L5 in Process Mining (2016;2ND ED) 
        out = miner._step_four(BOOK_LOG)
        self.assertEqual(out, BOOK_XL)

    def test_step_five(self):
        miner = AlphaMinerInstance()
        # following p173 and L5 in Process Mining (2016;2ND ED) 
        out = miner._step_five(BOOK_LOG, BOOK_XL)
        self.assertEqual(out, BOOK_YL) 

    def test_step_six(self):
        miner = AlphaMinerInstance()
        # following p173 and L5 in Process Mining (2016;2ND ED) 
        out = miner._step_six(BOOK_XL, BOOK_YL)
        self.assertEqual(out, BOOK_PL) 

    def test_step_seven(self):
        miner = AlphaMinerInstance()
        # following p173 and L5 in Process Mining (2016;2ND ED) 
        out = miner._step_seven(BOOK_LOG, BOOK_PL, BOOK_TI, BOOK_TO)
        self.assertEqual(out, BOOK_FL) 

    def test_mine(self):
        miner = AlphaMinerInstance()
        # following p173 and L5 in Process Mining (2016;2ND ED) 
        PL, TL, FL = miner.mine_model(BOOK_LOG)
        self.assertEqual(PL, BOOK_PL)
        self.assertEqual(TL, BOOK_TL_OUT)
        self.assertEqual(FL, BOOK_FL)


"""
This module provides functions to export out languages and different types 
of event logs constructs to the XES format.
"""
from pmkoalas import __version__

from pmkoalas.simple import EventLog
from pmkoalas.complex import ComplexEventLog
from pmkoalas._logging import debug,info, enable_logging, get_logger

from pmkoalas.xes_export import XesLogExtension, XesLogClassifier
from pmkoalas.xes_export import XES_LOG_TAG,XES_LOG_ATTRS
from pmkoalas.xes_export import XES_EXT_CONCEPT_NAME,XES_EXT_CONCEPT_PREFIX 
from pmkoalas.xes_export import XES_EXT_CONCEPT_URI

from pmkoalas.xes_export import XesString, XesInt, XesTime, XesFloat, XesBoolean
from pmkoalas.xes_export import XES_CONCEPT

from pmkoalas.xes_export import XesTrace,XesEvent

import os 
import logging
from typing import Union
from datetime import datetime

from xml.etree import ElementTree as ET

EXPORT_SIMPLE_TRACE_FORMAT = "trace {id:d}"

@enable_logging
def export_to_xes_simple(filepath:str, log:EventLog) -> None:
    """
    This exports a simple event log structure out into an XES format but 
    consider the following before using:
    - A simple event log does not consider time, so no time:timstamp 
       element will be produced.
    - Thus, the only attribute exported for events will be concept:name.
    - Traces will have a concept:name, and will be given a dummy 
       concept:name based on seen order from log.
    - For each trace variant, we add x number of traces, based on how many 
       times a variant is seen.
    - Eventlogs will have concept:name, using the name from the event log.
    - We do not assume that the filepath exists, and will create the parent 
      directory path if does not exist.
    """

    info(f"exporting log of size :: {len(log)}")

    # check filepath
    if (not os.path.exists(os.path.dirname(filepath))):
        os.makedirs(os.path.dirname(filepath),exist_ok=True)

        info(f"made directory for :: {filepath}")

    with open(filepath,"wb") as flog:
        # add log element
        xml_log = ET.Element( XES_LOG_TAG, XES_LOG_ATTRS)
        # make xml docuement
        xml_tree = ET.ElementTree(xml_log)
        # (1) add default extension for concept:name
        xml_log.append(XesLogExtension(XES_EXT_CONCEPT_NAME,
            XES_EXT_CONCEPT_PREFIX,
            XES_EXT_CONCEPT_URI)
        )
        # (2) add any globals
        # (3) add any classifiers
        xml_log.append(XesLogClassifier("activity-classifier",
         [XES_CONCEPT]))
        # (4) add any properties 
        # add concept:name to log
        xml_log.append(XesString(XES_CONCEPT, log.get_name()))
        # add meta_exporter to log as koalas
        xml_log.append(XesString("meta:exporter", "koalas"))
        xml_log.append(XesString("meta:exporter:version", 
         f"{__version__}"))
        # add traces
        trace_id = 1
        debug("starting trace conversion")
        for trace,count in log.__iter__():
            events = []
            # add a trace, count times
            for _ in range(count):
                # generate parent trace
                xml_trace = XesTrace(EXPORT_SIMPLE_TRACE_FORMAT.format(
                  id=trace_id))

                # only generate events once
                if (len(events) != len(trace)):
                  debug(f"Generating events for variant ::"+
                   f"{trace} x {count}")
                  # generate subelements
                  for act in trace.__iter__():
                      ev = XesEvent()
                      # add concept for event
                      ev.append(XesString(XES_CONCEPT, act))
                      # keep event
                      events.append(ev)  

                  if (get_logger().isEnabledFor(logging.DEBUG)):
                    debug(f"Generated sequence of events :: "+
                     f"{[ ET.tostring(e) for e in events ]}")

                # add events as subelements
                for ev in events:
                    xml_trace.append(ev) 

                # after adding all events to trace, add element to log
                xml_log.append(xml_trace)

                trace_id += 1
        debug(f"exported traces :: {trace_id-1}")
        # write out xml to file
        ET.indent(xml_tree, space="\t", level=0)
        xml_tree.write(flog, encoding="utf-8", method="xml",
         xml_declaration=True)
    
    info(f"exported log to :: {filepath}")

def create_xes_attribute(key:str, value:object) -> Union[XesString, XesInt, XesTime]:
    if isinstance(value, float):
       return XesFloat(key, value) 
    elif isinstance(value, bool):
       return XesBoolean(key, value)
    elif isinstance(value, int):
      return XesInt(key, value)
    elif isinstance(value, datetime):
      return XesTime(key, value)
    else:
      return XesString(key, value)


@enable_logging
def export_to_xes_complex(filepath:str, log:ComplexEventLog) -> None:
    """
    This exports a complex event log structure out into an XES format but 
    consider the following before using:
    - Traces will have a concept:name, and will be given a dummy 
       concept:name based on seen order from log if not presented.
    - Any data associated with an event, trace or event log will be added as 
      attributes at the associated level.
    - For each trace variant, we add all instances of that variant in seen order
    - Eventlogs will have concept:name, using the name from the event log.
    - We do not assume that the filepath exists, and will create the parent 
      directory path if does not exist.
    """

    if (isinstance(log, EventLog)):
        info("Changing to simple version as given log was simple")
        export_to_xes_simple(filepath, log)
        return 
    elif (not isinstance(log, ComplexEventLog)):
        raise ValueError(f"Was expecting a complex event log, but was given :: {type(log)}")

    info(f"exporting log of size :: {len(log)}")

    # check filepath
    if (not os.path.exists(os.path.dirname(filepath))):
        os.makedirs(os.path.dirname(filepath),exist_ok=True)

        info(f"made directory for :: {filepath}")

    with open(filepath,"wb") as flog:
        # add log element
        xml_log = ET.Element( XES_LOG_TAG, XES_LOG_ATTRS)
        # make xml docuement
        xml_tree = ET.ElementTree(xml_log)
        # (1) add default extension for concept:name
        xml_log.append(XesLogExtension(XES_EXT_CONCEPT_NAME,
            XES_EXT_CONCEPT_PREFIX,
            XES_EXT_CONCEPT_URI)
        )
        # (2) add any globals
        # (3) add any classifiers
        xml_log.append(XesLogClassifier("activity-classifier",
         [XES_CONCEPT]))
        # (4) add any properties 
        # add concept:name to log
        xml_log.append(XesString(XES_CONCEPT, log.get_name()))
        # add elements from data
        if (log.data() != None):
           for key,val in log.data().items():
              xml_log.append(create_xes_attribute(key,val))
        # add meta_exporter to log as koalas
        xml_log.append(XesString("meta:exporter", "koalas"))
        xml_log.append(XesString("meta:exporter:version", 
         f"{__version__}"))
        # add traces
        trace_id = 1
        debug("starting trace conversion")
        for _,complexes in log.get_instances().items():
            for trace in complexes:
              events = []
              # generate complex trace
              xml_trace = None
              if ( XES_CONCEPT in trace.data().keys()):
                name = trace.data()[XES_CONCEPT]
              else:
                name = EXPORT_SIMPLE_TRACE_FORMAT.format(
                  id=trace_id)
              xml_trace = XesTrace(name)
              # generate events
              debug(f"Generating events for variant ::"+
              f"{trace}")
              # generate subelements
              for cev in trace:
                  xml_ev = XesEvent()
                  # add concept for event
                  xml_ev.append(XesString(XES_CONCEPT, cev.activity()))
                  # add other attributes to event
                  for key,val in cev.data().items():
                     if key == XES_CONCEPT:
                        continue
                     else:
                        xml_ev.append(create_xes_attribute(key,val))
                  # keep event
                  events.append(xml_ev)  

              if (get_logger().isEnabledFor(logging.DEBUG)):
                debug(f"Generated sequence of events :: "+
                f"{[ ET.tostring(e) for e in events ]}")

              # add attributes to trace
              for key,val in trace.data().items():
                 if key == XES_CONCEPT:
                    continue
                 else:
                    xml_trace.append(create_xes_attribute(key,val))

              # add events as subelements
              for xml_ev in events:
                  xml_trace.append(xml_ev) 

              # after adding all events to trace, add element to log
              xml_log.append(xml_trace)
              trace_id += 1

        debug(f"exported traces :: {trace_id-1}")
        # write out xml to file
        ET.indent(xml_tree, space="\t", level=0)
        xml_tree.write(flog, encoding="utf-8", method="xml",
         xml_declaration=True)
    
    info(f"exported log to :: {filepath}")


"""
This module contains information about the structure of complex
api for process data, being how to represent an event,
a trace and an event log in complex form, which considers the
data perspective.
"""

from typing import Mapping, Iterable, Set, List, Tuple
from copy import deepcopy
from time import time

from pmkoalas._logging import info, debug, enable_logging
from pmkoalas.simple import Trace, EventLog

class ComplexEvent():
    """
    A complex form of an event, an atomic change in the state
    of a process.
    """

    STR_FORMAT = "(e : {act}|{map})"

    def __init__(self, activity:str, data:Mapping[str,object]) -> None:
        self._act = activity
        self._map = deepcopy(data)
        self._hash = hash(
            tuple(
                [self._act,]+
                [hash(tuple([key,val])) for key,val 
                 in self._map.items()]
                )
        )

    def activity(self) -> str:
        """ the process activity denoted by this event """
        return self._act

    def data(self) -> Mapping[str,object]:
        """ returns a copy of the data attached at this event """
        return deepcopy(self._map)    

    # data model functions
    def __getitem__(self, key):
        if key in self._map.keys():
            return deepcopy(self._map[key])
        else:
            return "UNDEFINED"
        
    def _pretty_map(self) -> str:
        str_map = "{"
        for key,val in self.data().items():
            str_map += f" '{key}' : {val},"
        return str_map[:-1] + "}"

    def __str__(self) -> str:
        label = self.STR_FORMAT.format(act=self._act,map=self._pretty_map())
        return label

    def __repr__(self) -> str:
        repr = "ComplexEvent(\n\t"
        repr += f"'{self._act}'" + ", {\n"
        for key,val in self._map.items():
            repr += f"\t'{key}' : {val.__repr__()},\n"
        repr += "\t}\n)"
        return repr
    
    def __hash__(self) -> int:
        return self._hash
    
class ComplexTrace():
    """
    A complex form of sequence of complex events, where a trace
    can have a invariant mapping attached for trace level attributes.
    """

    def __init__(self, events:Iterable[ComplexEvent], 
                 data: Mapping[str,object] = None) -> None:
        self._sequence = [ 
            deepcopy(event)
            for event
            in events
        ]
        self._len = len(self._sequence)
        if data == None:
            self._map = dict()
        elif isinstance(data,(dict)):
            self._map = deepcopy(data)
        else:
            raise ValueError(f"Given data is not a map/dict :: {type(data)}")
        self._hash = hash( 
            tuple(list(self._map.items()) + [ s.__hash__() for s in self._sequence])
        )
        self._acts = set([ s.activity() for s in self._sequence])

    # accessors
    def get_id(self) -> str:
        return "complex"
    
    def seen_activities(self) -> Set[str]:
        """ returns the activities seen in this trace """
        return deepcopy(self._acts)
    
    def data(self) -> Mapping[str,object]:
        """ returns the trace attributes """
        return deepcopy(self._map)
    
    def simplify(self) -> Trace:
        """ 
        returns a simplifed representation of this trace without data.
        """
        return Trace([ s.activity() for s in self])
    
    # data model functions
    def __str__(self) -> str:
        if (self._len < 1):
            return f"<>:{str(self._map)}"
        else:
            rep = "<"
            for event in self:
                rep = rep + f"{str(event)}, "
            rep = rep[:-2] + ">"
            return rep

    def __repr__(self) -> str:
        repr = "ComplexTrace(\n\t[\n" 
        # add events in sequence
        for ev in self:
            repr += "\t\t" + ev.__repr__() + ",\n"
        repr += "\t],\n"
        # add map
        repr += "\tdata= "+ self.data().__repr__() +"\n" 
        return repr + ")"
    
    def __iter__(self) -> Iterable[ComplexEvent]:
        for event in self._sequence:
            yield event 

    def __getitem__(self,no:int) -> ComplexEvent:
        if (isinstance(no,(int))):
            no = int(no)
            if (no < self._len and no >= 0):
                return self._sequence[no]
            else:
                raise ValueError("Invalid sequence reference, "+ 
                                 f"must be between 0 and {self._len}"+
                                 f" but was given {no}.")
        else:
            raise ValueError(
                "Invalid sequence reference, expecting a int"+
                f" but was given {type(no)}."      
            )

    def __eq__(self, __o: object) -> bool:
        if ( isinstance(__o, ComplexTrace )  ):
            if (self._map.__eq__(__o._map)): 
                if (self._len == __o._len):
                    for e,oe in zip(self, __o):
                        if (not e.__hash__() == oe.__hash__()):
                            return False 
                    return True
                else:
                    return False
            else:
                return False
        return False

    def __hash__(self) -> int:
        return self._hash

    def __len__(self) -> int:
        return self._len

DEFAULT_COMPLEX_LOG_NAME = "complex log"
class ComplexEventLog():
    """
    A complex collection of complex traces and can have an 
    invariant mapping. This collection can express several 
    languages.
    """

    @enable_logging
    def __init__(self, traces: Iterable[ComplexTrace],
                 data:Mapping[str,object] = None,
                 name:str = DEFAULT_COMPLEX_LOG_NAME) -> None:
        # middleman to multi set repr
        self._freqset = dict()
        self._instances = dict()
        self._len = 0
        self._variants = 0
        self._pop_size = 0
        self._acts = set([])
        self._start_acts = set([])
        self._end_acts = set([])
        self._traces = None
        self._map = deepcopy(data)
        info("Computing language...")
        start = time()
        for trace in traces:
            
            strace = trace.simplify()
            if (strace in self._instances):
                collector = self._instances[strace]
                self._pop_size += 1
                collector.append(trace)
            else:
                self._instances[strace] = list()
                self._instances[strace].append(trace)
                self._pop_size += 1
            trace = strace
            if (trace in self._freqset.keys()):
                self._freqset[trace] += 1
            else:
                self._acts = self._acts.union(
                    trace.seen_activities()
                )
                if (len(trace) > 0):
                    self._start_acts.add(trace[0])
                    self._end_acts.add(trace[-1])
                self._freqset[trace] = 1
                self._variants += 1
            self._len += 1
        self._traces = set([ t for t in self._freqset.keys() ])
        info(f"Computed language in {(time()-start)*1000:.0f}ms")
        self.name = name 

    @enable_logging
    def simplify(self) -> EventLog:
        """
        Simplifies this event log to only consider the process
        activities. Returns a new instance of simple event log.
        """
        simple_traces = []
        for trace, freq in self._freqset.items():
            simple_traces = simple_traces + ([ trace ] * freq)
        return EventLog(simple_traces, self.name)

    def seen_activities(self) -> Set[str]:
        "Get a language of process activities from this language."
        return deepcopy(self._acts)

    def seen_start_activities(self) -> Set[str]:
        "Get a set of start activities from this language."
        return deepcopy(self._start_acts)

    def seen_end_activities(self) -> Set[str]:
        "Get a set of end activities from this language."
        return deepcopy(self._end_acts)
    
    def simple_language(self) -> Set[Trace]:
        "Get a simplified trace language from this language."
        return set(list(self._freqset.keys()))
    
    def simple_stochastic_language(self) -> Mapping[Trace,float]:
        "Get a simplified stochastic language from this language."
        return self._freqset.copy()
    
    def get_instances(self) -> Mapping[Trace, List[ComplexTrace]]:
        """ 
        Get a map between seen simple traces and instances 
        of complex traces.
        """
        return deepcopy(self._instances)
    
    def seen_instances_for(self, trace:Trace) -> Set[ComplexTrace]:
        """
        Explores this collection for instances of the given 
        simplified trace, may return instances or an empty collection.
        """
        if (trace in self._traces):
            return deepcopy(self._instances[trace]) 
        else:
            return []
        
    def get_name(self) -> str:
        " returns the name of this collection."
        return self.name

    def get_nvariants(self) -> int:
        " returns the number of trace variants in this collection."
        return self._variants
    
    def get_population_size(self) -> int:
        " returns the number of traces instances in this collection."
        return self._pop_size
    
    def data(self) -> Mapping[str,object]:
        " returns a data mapping for this collection, i.e. log-level attributes."
        return self._map
    
    # data model functions
    def __len__(self) -> int:
        return self.get_population_size()
    
    def __iter__(self) -> Iterable[Tuple[Trace,List[ComplexTrace]]]:
        for strace, collector in self._instances.items():
            yield deepcopy(strace), deepcopy(collector)

    def __str__(self) -> str:
        if (self._variants < 1):
            return "[]"

        _str = "[\n"
        for strace,complexes in self:
            for complex in complexes:
                _str = _str + "\t" + str(complex)+":"+str(complex.data())+",\n"
        _str = _str[:-2] + "\n]"
        return _str
    
    def __repr__(self) -> str:
        repr = "ComplexEventLog(\n\t[\n"
        # add complex traces
        for _, complexes in self:
            for complex in complexes:
                repr = repr + "\t" + complex.__repr__() +",\n"
        repr = repr +"\t],\n"
        # add data for event log
        repr = repr + "\tdata="+self._map.__repr__()+"\n"
        return repr + ")"
    
    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, ComplexEventLog)):
            if (__o.get_nvariants() == self.get_nvariants()):
                if (__o.get_population_size() == self.get_population_size()):
                    return self.get_instances() == __o.get_instances()  
        elif (isinstance(__o, EventLog)):
            return self.simplify() == __o
        # fall through
        return False




import unittest
from pmkoalas.simple import Trace,EventLog
from pmkoalas.dtlog import convert

class TraceTest(unittest.TestCase):
    def test_init(self):
        t = Trace(['a','c'])
 
    def test_str(self):
        self.assertEqual( '<b>', str(Trace(['b']) ) )
        self.assertEqual( '<a,b>', str(Trace(['a','b']) ) )
        self.assertEqual( '<>', str(Trace([]) ) )

    def test_trace_activities(self):
        t = Trace(['a','b','c'])
        self.assertEqual(t.seen_activities(), set(['a','b','c']))
        t = Trace([])
        self.assertEqual(t.seen_activities(), set([]))

    def test_lang_activities(self):
        log = convert("a b c", "a b", "c", "")
        self.assertEqual(log.seen_activities(), 
            set(['a','b','c'])
        )
        log = EventLog([])
        self.assertEqual(log.seen_activities(), set())

    def test_lang_start_activities(self):
        log = convert("a b c", "b c", "c")
        self.assertEqual(log.seen_start_activities(),
            set(['a', 'b', 'c'])
        )
        log = convert("a b c", "a b", "a b c")
        self.assertEqual(log.seen_start_activities(), 
            set(['a'])
        )

    def test_lang_end_activities(self):
        log = convert("a b c", "b c", "c")
        self.assertEqual(log.seen_end_activities(),
            set(["c"])
        )
        log = convert("a b c", "a b", "a", "d")
        self.assertEqual(log.seen_end_activities(),
            set(['c','b','a', 'd'])
        )

if __name__ == '__main__':
    unittest.main()




"""
This module contains information about the structure of the 
simplified api for process data, being how to represent an event, 
a trace and an event log.
"""

from typing import Iterable, List, Mapping, Set, Tuple
from copy import deepcopy
from time import time

from pmkoalas._logging import info, debug, enable_logging
from pmkoalas.directly import DirectlyFollowPair,FollowLanguage
from pmkoalas.directly import DIRECTLY_SOURCE,DIRECTLY_END

class Trace():
    """
    A simplified representation of a sequence of events
    """

    def __init__(self, sequence: List[str]) -> None:
        self.sequence = deepcopy(sequence)
        self._len = len(self.sequence)
        self._hash = hash(tuple(event for event in self.sequence))
        self._acts = set(self.sequence)
    
    # accessors
    def get_id(self) -> str:
        return self.id

    def seen_activities(self) -> Set[str]:
        return self._acts

    # data model functions
    def __str__(self) -> str:
        if len(self.sequence) == 0:
            return "<>"
        rep = "<"
        for event in self:
            rep = rep + f"{str(event)},"
        rep = rep[:-1] + ">"
        return rep

    def __repr__(self) -> str:
        event_repr = "["
        for event in self:
            event_repr = event_repr + event.__repr__() +","
        event_repr = event_repr[:-1] + "]"

        return f"Trace({event_repr})"

    def __iter__(self) -> Iterable[str]:
        for event in self.sequence:
            yield event

    def __getitem__(self,key:int) -> str:
        return self.sequence[key]

    def __eq__(self, __o: object) -> bool:
        if ( isinstance(__o, Trace )  ):
            return self.sequence.__eq__(__o.sequence)
        return False

    def __hash__(self) -> int:
        return self._hash

    def __len__(self) -> int:
        return self._len

DEFAULT_SIMPLE_LOG_NAME="simple"

class EventLog():
    """
    A simplified collection of traces, which can express
    several languages.
    """

    @enable_logging
    def __init__(self, traces: Iterable[Trace], 
                 name:str=DEFAULT_SIMPLE_LOG_NAME) -> None:
        # middleman to multi set repr
        self._freqset = dict()
        self._len = 0
        self._variants = 0
        self._acts = set([])
        self._start_acts = set([])
        self._end_acts = set([])
        self._traces = None
        info("Computing language...")
        start = time()
        for trace in traces:
            if (trace in self._freqset.keys()):
                self._freqset[trace] += 1
            else:
                self._acts = self._acts.union(
                    trace.seen_activities()
                )
                if (len(trace) > 0):
                    self._start_acts.add(trace[0])
                    self._end_acts.add(trace[-1])
                self._freqset[trace] = 1
                self._variants += 1
            self._len += 1
        self._traces = set([ t for t in self._freqset.keys() ])
        info(f"Computed language in {(time()-start)*1000:.0f}ms")
        self.name = name 
        self._relations = None

    def seen_activities(self) -> Set[str]:
        "Get a language of process activities from this language"
        return deepcopy(self._acts)

    def seen_start_activities(self) -> Set[str]:
        "Get a set of start activities from this language"
        return deepcopy(self._start_acts)

    def seen_end_activities(self) -> Set[str]:
        "Get a set of end activities from this language"
        return deepcopy(self._end_acts)

    def language(self) -> Set[Trace]:
        "Get a trace language from this language"
        return set(list(self._freqset.keys()))
    
    def stochastic_language(self) -> Mapping[Trace,float]:
        "Get a stochastic language from this language"
        return self._freqset.copy()

    @enable_logging
    def directly_follow_relations(self) -> FollowLanguage:
        "Get the directly flow relations for this language"
        if (self._relations == None):
            # compute directly flow relations
            self._relations = FollowLanguage([])
            start = time()
            computed_times = []
            info("Starting computation of relations")
            for tid,(trace,freq) in enumerate(self._freqset.items()):
                if (len(trace) < 1):
                    continue
                compute_time = time()
                relations = []
                # build initial flow
                debug(f"{trace=} @ {freq=}")
                debug(f"{DIRECTLY_SOURCE} to {trace[0]=} @ {freq}")
                relations.append(DirectlyFollowPair(left=DIRECTLY_SOURCE, 
                 right=trace[0], freq=freq))

                # build body flows
                for src,trg in zip(trace[:-1],trace[1:]):
                    debug(f"{src=} to {trg=} @ {freq=}")
                    relations.append(DirectlyFollowPair(left=src,
                     right=trg, freq=freq))

                # build exit flow 
                debug(f"{trace[-1]=} to {DIRECTLY_END} @ {freq}")
                relations.append(DirectlyFollowPair(left=trace[-1], 
                 right=DIRECTLY_END, freq=freq))

                # update lang
                self._relations  = self._relations + \
                    FollowLanguage(relations)
                
                if (len(computed_times) < 30):
                    computed_times.append(time() - compute_time)
                else:
                    computed_times.pop(0)
                    computed_times.append(time() - compute_time)

                if (tid > 0 and (tid % 100) == 0):
                    avg_time = (sum(computed_times)/ 
                     len(computed_times))* 1000
                    info(f"computed {tid}/{self._variants} @ " + 
                     f"{avg_time:.1f}ms/variant"
                    )

            # relations computed 
            info(f"Computed relations in {(time()-start)*1000:.0f}ms")
            return self._relations
        else:
            info("Already computed relations, returning existing" + 
             "computation.")
            return self._relations

    # data model functions
    def __len__(self) -> int:
        return self._len

    def __iter__(self) -> Iterable[Tuple[Trace,int]]:
        return self._freqset.items().__iter__()

    def __str__(self) -> str:
        if (self._variants < 1):
            return "[]"

        _str = "["
        for trace,freq in self:
            _str = _str + str(trace) + f"^{freq},"
        _str = _str[:-1] + "]"
        return _str

    def __repr__(self) -> str:
        rep = "EventLog(\n"
        for trace, freq in self:
            rep += "\t[" + trace.__repr__() + f"] * {freq}+\n"
        return rep[:-2] + "\n)"

    # accessors 
    def get_name(self) -> str:
        return self.name

    def get_nvariants(self) -> int:
        return self._variants

    # membership test
    def __contains__(self, other):
        if isinstance(other, Trace):
            return other in self._freqset.keys()
        raise NotImplemented("Membership test not defined for :: "
         + type(other))

    # Rich comparisons 
    # https://peps.python.org/pep-0207/#classes
    # ==, <, <=, >, >=
    def __eq__(self,other) -> bool:
        if isinstance(other,EventLog):
            return self.stochastic_language() == \
             other.stochastic_language()
        return False

    def __lt__(self,other) -> bool:
        """
        Tests whether this event log is a proper subset of 
        another event log.
        """
        if isinstance(other,EventLog):
            this_traces = self._traces
            other_traces = other._traces
            if (this_traces.issubset(other_traces)):
                diff_set = other_traces.difference(
                    this_traces )
                return len(diff_set) > 0
            return False
        raise NotImplementedError("Subset comparsion" +
         f" between an EventLog (simple) and {type(other)}" +
          " is not defined.") 

    def __le__(self,other) -> bool:
        """
        Tests whether this event log is a subset of 
        another event log.
        """ 
        if isinstance(other,EventLog):
            this_traces = self._traces
            other_traces = other._traces
            return this_traces.issubset(other_traces)
        raise NotImplementedError("Subset comparsion" +
         f" between an EventLog (simple) and {type(other)}" +
          " is not defined.") 

    def __gt__(self,other) -> bool:
        pass 

    def __ge__(self,other) -> bool:
        pass 
    
    # Emulating numeric types
    # do we need these? could be cool for extra bits
    # <<, >>, &, ^, |
    def __lshift__(self, other):
        pass 

    def __rshift__(self, other):
        pass 

    def __and__(self, other):
        pass

    def __xor__(self, other):
        pass

    def __or__(self, other):
        pass






import unittest
from pmkoalas.dtlog import *

class DTLogTest(unittest.TestCase):
    def test_convert_empty(self):
        self.assertEqual( EventLog([]), convert() )

    def test_convert_single_trace(self):
        self.assertEqual( Trace(['a']), convertTrace("a" ) )

    def test_convert_single_trace_multi_event(self):
        self.assertEqual( Trace(['a','b']), convertTrace("a b" ) )
        self.assertEqual( Trace(['d','a','b']), convertTrace("d a b" ) )
        self.assertEqual( Trace(['jill','alex','thingy']), 
                          convertTrace("jill alex thingy" ) )
        self.assertEqual( Trace(['a','b']), 
                          convertTrace("a-b", delimiter='-' ) )

    def test_convert_single(self):
        self.assertEqual( EventLog([Trace(['a'])]), convert("a" ) )
        self.assertEqual( EventLog([Trace(['a','b'])]), convert("a b" ) )

    def test_convert_triple(self):
        self.assertEqual( EventLog([Trace(['a','b']),
                                    Trace(['a']), 
                                    Trace(['a','b'])]), 
                          convert("a b","a","a b" ) )



if __name__ == '__main__':
    unittest.main()





"""
Module to control logging and logger instance so not to clash with other loggers
"""

import logging
from logging import Logger
from functools import wraps

LOG_FRMT = '%(asctime)s|%(filename)-18s|%(funcName)-25s|%(levelname)-8s|: %(message)s'

def get_logger() -> Logger:
    """
    This will get/create the unique logger for koalas.
    """

    # request or create a logger
    logger = logging.getLogger("koalas-log")

    # only do setup if needed
    if (not logger.hasHandlers()):
        logger.setLevel(logging.ERROR)
        fmt = LOG_FRMT
        fmt_date = '%Y-%m-%dT%T'
        formatter = logging.Formatter(fmt, fmt_date)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger 

def info(msg:str, *args, **kwargs):
    """
    Sends a info message to the koalas logger.
    """
    get_logger().info(msg, stacklevel=2, *args, **kwargs) 

def debug(msg:str, *args, **kwargs):
    """
    Sends a debug message to the koalas logger.
    """
    get_logger().debug(msg, stacklevel=2, *args, **kwargs) 

def warn(msg:str, *args, **kwargs):
    """
    Sends a warning message to the koalas logger.
    """
    get_logger().warn(msg, stacklevel=2, *args, **kwargs) 

def error(msg:str, *args, **kwargs):
    """
    Sends a error message to the koalas logger.
    """
    get_logger().error(msg, stacklevel=2, *args, **kwargs)

def setLevel(level):
    """
    Sets the logging level for koalas.
    """
    get_logger().setLevel(level)

def enable_logging(func):
    """
    Optional Parameters
    -------------------
    - debug [`True/False`]: if `True` function will show info level msgs
    - debug_level: sets the level of logging to show. 
    """
    # add to docstrings so 
    if (func.__doc__ != None):
        func.__doc__  += enable_logging.__doc__

    @wraps(func)
    def wrapped(*args, **kwargs):
        # a tuple of the names of the parameters that func accepts
        func_params = func.__code__.co_varnames[:func.__code__.co_argcount]
        # grab all of the kwargs that are not accepted by func
        extra = set(kwargs.keys()) - set(func_params)
        debug = kwargs.pop("debug",False)
        level = kwargs.pop("debug_level", logging.INFO)
        # set logger settings
        if debug:
            # set debug on at info level
            get_logger().setLevel(level)
        # run function as intended
        val = func(*args, **kwargs)
        # reset logger
        get_logger().setLevel(logging.ERROR)
        # pass value back if needed
        return val
    
    return wrapped


import unittest
from os import path 
from datetime import datetime

from pmkoalas.read import read_xes_simple,read_xes_complex
from pmkoalas.simple import EventLog,Trace
from pmkoalas.complex import ComplexEventLog, ComplexTrace, ComplexEvent

SSMALL = path.join(".","tests","small_01.xes")
WSMALL = path.join(".","tests","small_02.xes")
OSMALL = path.join(".","tests","small_03.xes")
DSMALL = path.join(".","tests","small_04.xes")

class DTLogTest(unittest.TestCase):

    def test_successful_read(self):
        try:
            log = read_xes_complex(SSMALL) 
        except:
            self.fail("Failed to parse log")

        real_log = EventLog( [
                Trace(["A","B","C","D","E"]),
                Trace(["A","B","C","D","E"]),
                Trace(["A","B","C","D","D"]),
                Trace(["A","A","A","B","E"]),
                Trace(["A","B","C","E","E"]),
            ])
        self.assertEqual(log, real_log)

    def test_unsucessful_read(self):
        try:
            log = read_xes_complex(WSMALL)
        except:
            return
        self.fail()

    def test_unsuccessful_label_change(self):
        try:
            log = read_xes_complex(SSMALL, label_attribute="foo:bah") 
        except:
            return 
        self.fail("Should fail when label attribute is not found!")

    def test_label_change(self):
        try:
            log = read_xes_complex(OSMALL, label_attribute="name") 
        except:
            self.fail("Failed to parse log after changing label attribute!")

        real_log = EventLog( [
                Trace(["E","D","C","B","A"]),
                Trace(["E","D","C","B","A"]),
                Trace(["D","D","C","B","A"]),
                Trace(["E","B","A","A","A"]),
                Trace(["E","E","C","B","A"]),
            ])
        self.assertEqual(log, real_log)

    def test_data_mappings(self):
        try:
            log = read_xes_complex(DSMALL) 
        except:
            self.fail("Failed to parse log")

        real_log = ComplexEventLog(
            [ 
                ComplexTrace( 
                    [ 
                        ComplexEvent(
                            'A',
                            {
                                'time:timestamp' : datetime.fromisoformat("2122-01-01T01:00:00.000+10:00"),
                                'res' : 0,
                                'life' : False,
                            }
                        ),
                        ComplexEvent(
                            'B',
                            {
                                'time:timestamp' : datetime.fromisoformat("2122-01-01T02:00:00.000+10:00"),
                                'res' : 1,
                                'cost:life' : 0.246,
                                'life' : False,
                            }
                        ),
                        ComplexEvent(
                            'C',
                            {
                                'time:timestamp' : datetime.fromisoformat("2122-01-01T03:00:00.000+10:00"),
                                'res' : 1,
                                'life' : True
                            }
                        )
                    ],
                    data = { 'concept:name' : 'trace_01', 'trace:cost' : 10}
                )
            ],
            data= {'concept:name' : 'A simple complex log',
                   'extracted' : datetime.fromisoformat("2122-01-05T01:00:00.000+10:00"),
                   'exporter:ver' : 1,
                   'percent:life' : 0.754,
                   'life' : True }
        )
        self.maxDiff = None
        self.assertEqual(log.__repr__(), real_log.__repr__())

    def nottest_read_bad(self):
        with self.failUnlessRaises(Exception): 
            log = read_xes_complex(path.join(".","tests","small_bad.xes")) 
            print(f"{log}")

if __name__ == '__main__':
    unittest.main()




"""
This modules handles creating directly flows pairs from a 
language.
"""
from tqdm import tqdm

from copy import deepcopy
from time import time
from typing import Iterable,Dict,List,Set
from tempfile import TemporaryFile
from random import randint

from pmkoalas._logging import enable_logging,info,debug
DIRECTLY_SOURCE = "SOURCE"
DIRECTLY_END = "END"

class DirectlyFollowPair():
    """
    This class describes one directly flow relation between
    two process activities seen in a language.
    """

    def __init__(self,left:str,right:str,freq:int) -> None:
        self._left = left.replace("\n","") 
        self._right = right.replace("\n","")
        self._hash = hash((self._left,self._right))
        self._freq = freq

    def left(self) -> str:
        "Source process activity"
        return self._left
    
    def right(self) -> str:
        "Target process activity"
        return self._right

    def frequency(self) -> int:
        "Support for flow relation"
        return self._freq

    def incre(self,count:int=1) -> None:
        "Increase support for relation"
        self._freq += count 

    def decre(self,count:int=1) -> None:
        "Decrease support for relation"
        self._freq -= count
    
    def copy(self) -> 'DirectlyFollowPair':
        return DirectlyFollowPair(self._left, self._right, self._freq)

    # data model functions
    def __str__(self) -> str:
        return f"({self._left} -> {self._right})^{self._freq}"

    def __repr__(self) -> str:
        left = self._left.replace("'","\\'")
        right = self._right.replace("'","\\'")
        return f"DirectlyFlowsPair(left='{left}'"+ \
            f",right='{right}',freq={self._freq})"

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, DirectlyFollowPair)):
            return self.__hash__() == __o.__hash__()
        return False

class DirectlyFollowWalk():
    """
    A walk of FlowLanguage, which describes the absolute power of the walk.
    """

    def __init__(self,start:DirectlyFollowPair) -> None:
        self._pow = start.frequency()
        self._loops = []
        self._nodes = [start.left(), start.right()]
        self._walk = [start.copy()]
        self._hash = hash(tuple([0]))
        self._hash = self._compute_hash()

    def append(self, next:DirectlyFollowPair) -> object:
        if (next.left() != self._walk[-1].right()):
            raise ValueError("Given pair does not connect to last.")
        if (next.right() in self._nodes):
            self._loops.append(next.copy())
        self._pow += next.frequency()
        self._walk.append(next.copy())
        self._nodes.append(next.right())
        self._hash = self._compute_hash()
        return self

    def power(self) -> float:
        "Return how frequent this walk is."
        return self._pow

    def get(self) -> List[DirectlyFollowPair]:
        "Return the connected walks of pairs"
        return deepcopy(self._walk)

    def last(self) -> DirectlyFollowPair:
        "Returns the last seen pair in the walk."
        return self._walk[-1].copy()

    def should_cross(self, next:DirectlyFollowPair) -> bool:
        "Returns if the walks should cross this path."
        if (next in self._loops):
            return False 
        return len([ p for p in self._walk if p == next]) <= 2

    def check_membership(self, checkfor:DirectlyFollowPair) -> bool:
        "Check if this walk has crossed this pair before."
        return checkfor in self._walk

    def copy(self):
        "Returns a new instance of this walk"
        return deepcopy(self)

    def convert_to_trace(self) -> object:
        "Converts this walk into a simple trace."
        from pmkoalas.simple import Trace
        return Trace([ step.right() for step in self._walk[:-1] ])

    def _compute_hash(self):
        return hash(tuple(
            [ 
                self._hash,
                self._walk[-1].__hash__(),
                self._pow
            ]
        ))

    # data model functions
    def __str__(self) -> str:
        repr = f"{self._walk[0].left()} -> {self._walk[0].right()}"
        for step  in self._walk[1:]:
            repr += f" -> {step.right()}"
        return repr

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, DirectlyFollowWalk)):
            return self.__hash__() == __o.__hash__()
        return False

    def __repr__(self) -> str:
        repr = f"DirectlyFlowWalk({self._walk[0].__repr__()})"
        for pair in self._walk[1:]:
            repr += f".append({pair.__repr__()})"
        return repr 

class FollowLanguage():
    """
    A language of directly follows relations.
    """

    def __init__(self, pairs:Iterable[DirectlyFollowPair]) -> None:
        self._relations = {}
        self._starts = {}
        self._ends = {}
        self._froms = {}
        self._tos = {}
        self._activities = set()
        self._walks = None
        self._pairs = 0
        start = time()
        debug("Computing new flow language")
        for pair in pairs:
            self._activities.add(pair.left())
            self._activities.add(pair.right())
            # add to general collection
            self._update_state(pair,self._relations, "collection")
            # add to starts if needed
            if (pair.left() == DIRECTLY_SOURCE):
                self._update_state(pair,self._starts, "starts")
            # add to ends if needed
            if (pair.right() == DIRECTLY_END):
                self._update_state(pair,self._ends, "ends")
        self._pairs = len(self._relations.items())
        debug("Computed new flow language in" +
         f" {(time() - start)*1000:.1f}ms")
        debug(f"Computed :: {str(self)}")

    def _introduce_pairs(self, pairs: List[DirectlyFollowPair]):
        """
        Internal function to update a state spaces with new paris.
        """
        start = time()
        debug("Computing new flow language")
        for pair in pairs:
            self._activities.add(pair.left())
            self._activities.add(pair.right())
            # add to general collection
            self._update_state(pair,self._relations, "collection")
            # add to starts if needed
            if (pair.left() == DIRECTLY_SOURCE):
                self._update_state(pair,self._starts, "starts")
            # add to ends if needed
            if (pair.right() == DIRECTLY_END):
                self._update_state(pair,self._ends, "ends")
        self._pairs = len(self._relations.items())
        debug(f"Computed new flow language in {(time() - start)*1000:.1f}ms")
        debug(f"Computed :: {str(self)}")

    def _update_state(self, pair:DirectlyFollowPair, state:Dict,
        state_name:str):
        """
        Internal function to update a state space with a pair.
        A pair may be used differently in each state space, so
        always take a copy.
        """
        val = state.get(pair,None)
        # check if flow already is recorded
        debug(f"{state_name} :: existing flow? {str(val)=}")
        if (val != None):
            newval = val.copy()
            newval.incre(pair.frequency())
            debug(f"{state_name} :: update : {val}")
            state[newval] = newval
        else:
            debug(f"{state_name} :: added : {pair}")
            state[pair] = pair.copy()

    def starts(self) -> List[DirectlyFollowPair]:
        "Returns all starting directly flow pairs."
        return list(self._starts.values())

    def ends(self) -> List[DirectlyFollowPair]:
        "Returns all ending directly flow pairs."
        return list(self._ends.values())

    def get(self, target:str) -> List[DirectlyFollowPair]:
        "Returns all pairs with left as target."
        relations = list()
        for pair in self:
            if (pair.left() == target):
                relations.append(pair)
        return relations

    def contains(self, pair:DirectlyFollowPair) -> bool:
        "Checks if pair is found in this language"
        return pair in self._relations.keys()

    def find(self, pair:DirectlyFollowPair) -> DirectlyFollowPair:
        "Finds the equivalent pair in this language"
        if (not self.contains(pair)):
            raise ValueError("pair not found in language")
        else:
            return self._relations[pair].copy()

    @enable_logging
    def approx_walks(self, attempts:int=10000) -> List[DirectlyFollowWalk]:
        """
        Returns approximates walks from flow language based on frequency.
        Return list of walks will be unique but may not cover all directly flows
        relations. 
        """
        finished = set()
        starts = self.starts()
        ends= self.ends()
        starts.sort(key=lambda x: x.frequency())
        sfreq = [ (s,sum([ ss.frequency() for ss in starts[:i]])) 
         for i,s in enumerate(starts) ]
        info("approximating walks...")
        start = time()
        for attempt in range(attempts):
            # coin flip on starting point
            flip = randint(0, sfreq[-1][-1])
            walk = None
            for starting,chance in sfreq:
                if (flip <= chance):
                    walk = DirectlyFollowWalk(starting)
                    break
            failed = False
            # begin flipping and choosing
            while (not failed and walk.last() not in ends):
                # get next steps
                poss = [ pair for pair in self.get(walk.last().right()) 
                     if walk.should_cross(pair) ]
                if (len(poss) < 1):
                    failed = True
                    continue
                # flip between choices
                poss.sort(key=lambda x: x.frequency())
                nextsfreq = [ 
                    (s,sum([ss.frequency() for ss in poss[:i]])) for i,s 
                    in enumerate(poss)
                ]
                # 
                flip = randint(0,nextsfreq[-1][-1])
                for succ,chance in nextsfreq:
                    if (flip <= chance):
                        walk = walk.append(succ)
                        break

            # add to approximates
            if (not failed):
                finished.add(walk)

            if attempt > 0 and (attempt % 1000) == 0:
                info(f"Completed {attempt}/{attempts} attempts...")

        info(f"finished approximation in {(time() - start)*1000:.1f}ms")
        info(f"found {len(finished)} walks while approximating...")
        return finished


    @enable_logging
    def walks(self) -> List[DirectlyFollowWalk]:
        """Returns all walks possible in the language. Warning may not finish 
        compute in finite time or in finite memory. """
        # check computation
        if (self._walks == None):
            info("computing walks...")
            start = time()
            # init ques 
            finished = []
            seen = SetArray()
            que = QuequeContainer() 
            deadlocks = False
            deadlock_count = 0
            # add starting points
            for pair in self.starts():
                walker = DirectlyFollowWalk(pair)
                que.append(walker)
            # iterative build walks
            info(f"starting que size {len(que)}...")
            while len(que) > 0:
                nque = QuequeContainer()
                for walk in tqdm(que):
                    # is walk at the end of the language
                    if (walk.last() in self.ends()):
                        finished.append(walk)
                        continue
                    # get possible pairs to expand to
                    poss = [ pair for pair in self.get(walk.last().right()) 
                     if walk.should_cross(pair) ]
                    if (len(poss) < 1):
                        deadlocks = True
                        deadlock_count += 1
                        debug("closing walk before reaching end :: " + str(walk))
                    # create new walks
                    for next_pair in poss:
                        nwalk = walk.copy()
                        nwalk.append(next_pair)
                        if (nwalk.__hash__() not in seen):
                            nque.append(nwalk)
                            seen.add(nwalk.__hash__())
                # start next cycle
                que = nque
                info(f"cycled que size {len(que)}...")
            # save computed walks
            info(f"finished computing a toal of {len(finished)} walks...")
            info(f"computation took {(time() -start)*1000:.1f}ms")
            if (deadlocks):
                info(f"numbder of deadlocks seen :: {deadlock_count}")
            else:
                info("No deadlocks occured during computation")
            finished.sort(reverse=True,key=lambda x: x.power())
            self._walks = finished
        return self._walks
    
    def activities(self) -> Set[str]:
        "Returns all activities seen in the language"
        return self._activities - set([DIRECTLY_END,DIRECTLY_SOURCE])


    # data model functions
    def __add__(self, other:object) -> object:
        if (isinstance(other, FollowLanguage)):
            new_flang = deepcopy(self)
            new_flang._introduce_pairs(other._relations.values())
            return new_flang
        raise NotImplemented("Flow language addition not support with" +\
             f" :: {type(other)}")

    def __iter__(self) -> Iterable[DirectlyFollowPair]:
        for pair in self._relations.values():
            yield pair

    def __len__(self) -> int:
        return self._pairs

    def __str__(self) -> str:
        rep = ""
        for pair in self._relations.values():
            rep += str(pair)+", "
        return "[ " + rep[:-2] + " ]"

    def __repr__(self) -> str:
        repr = "FlowLanguage([\n\t"
        for pair in self._relations.values():
            repr += f"{str(pair.__repr__())},\n\t"
        return repr[:-2] + "\n])"

class SetArray():
    """
    An alternative storage instances for large sets. Only defines add and 
    membership.
    """

    def __init__(self,tmpsize=100000) -> None:
        self._members = [set()]
        self._max_size = tmpsize
        self._curr_size = 0 
        self._curr_set = self._members[-1]

    def add(self, item):
        if (item not in self):
            self._curr_size += 1
            self._curr_set.add(item)
            if (self._curr_size >= self._max_size):
                self._members.append(set())
                self._curr_set = self._members[-1]
                self._curr_size = 0

    def __contains__(self, item) -> bool:
        check = False 
        for setter in self._members:
            check = check or item in setter 
        return check


class QuequeContainer():
    """
    A alternative storage of things, with a set size of items to be stored in
    memory. Expects that items have a accurate representation via __repr__.
    """

    def __init__(self, hash:bool=False, tmpsize=10000) -> None:
        self._temp_hdrive = TemporaryFile()
        self._use_hashes = hash
        self._temp_lines = 0
        self._tmpsize = tmpsize
        self._storage = [ None for _ in range(self._tmpsize)]
        self._storage_used = 0

    def append(self, item:object):
        # if there is a slot free in memory then use it
        if (self._storage_used < self._tmpsize):
            self._storage[self._storage_used] = item 
            self._storage_used += 1
        else: 
            # otherwise flush content to file
            self._temp_hdrive.write(b'[')
            for sitem in self._storage:
                if (self._use_hashes):
                    self._temp_hdrive.write(sitem.__hash__().encode("UTF-8")) 
                else:
                    self._temp_hdrive.write(sitem.__repr__().encode("UTF-8"))
                self._temp_hdrive.write(b',')
                self._temp_lines += 1
            self._temp_hdrive.write(b']\n')
            self._temp_hdrive.flush()
            # then repopulate slots in memory
            self._storage_used = 0
            self._storage = [ None for _ in range(self._tmpsize)]
            self._storage[self._storage_used] = item 
            self._storage_used += 1
            
    
    def __iter__(self) -> Iterable[object]:
        # loop through storage
        for item in self._storage[:self._storage_used]:
            yield item 
        # loop through dumps in file if need
        if (self._temp_lines > 0):
            # flush file and return to start
            self._temp_hdrive.flush()
            old_index = self._temp_hdrive.tell()
            self._temp_hdrive.seek(0)
            # get a block and compute full list
            for bytes in self._temp_hdrive:
                items = eval(bytes)
                # yield item from block
                for item in items:
                    yield item
            # reset file seek in case of further appending
            self._temp_hdrive.seek(old_index)

    def __len__(self) -> int :
        return self._storage_used + self._temp_lines


import unittest
from os import path 
from tempfile import TemporaryDirectory

from pmkoalas.read import read_xes_simple
from pmkoalas.export import export_to_xes_simple

import xmlschema

SSMALL = path.join(".","tests","small_01.xes")
WSMALL = path.join(".","tests","small_02.xes")
OSMALL = path.join(".","tests","small_03.xes")

class DTLogTest(unittest.TestCase):

    def test_successful_export(self):
        try:
            log = read_xes_simple(SSMALL) 
        except:
            self.fail("Failed to parse log")
        fdir = TemporaryDirectory()
        try:
            filepath = path.join(fdir.name, "test_log")
            export_to_xes_simple(filepath, log) 
        except:
            self.fail("Failed to export log")
        finally:
            fdir.cleanup()

    def test_equals_export_import(self):
        try:
            log = read_xes_simple(SSMALL) 
        except:
            self.fail("Failed to parse log")
        fdir = TemporaryDirectory()
        try:
            filepath = path.join(fdir.name, "test_log")
            export_to_xes_simple(filepath, log) 
        except:
            fdir.cleanup()
            self.fail("Failed to export log")
            return
        
        log2 = read_xes_simple(filepath)
        self.assertEqual(log, log2)

        fdir.cleanup()

    def test_xml_valid(self):
        try:
            log = read_xes_simple(SSMALL) 
        except:
            self.fail("Failed to parse log")

        fdir = TemporaryDirectory()
        try:
            filepath = path.join(fdir.name, "test_log")
            export_to_xes_simple(filepath, log) 
        except:
            fdir.cleanup()
            self.fail("Failed to export log")
            return

        schema = xmlschema.XMLSchema(
            path.join(".","tests","xes-ieee-1849-2016.xsd"),
            namespace="http://www.xes-standard.org/"
        )
        try:
            schema.validate(
                filepath, 
                namespaces={ 'xes' : "http://www.xes-standard.org/"},
            )
            fdir.cleanup()
            return 
        except Exception as e: 
            fdir.cleanup()
            self.fail("Failed to validate exported xes")


"""
Parses delimited traces into koalas EventLog objects.
"""
from typing import Iterable
from pmkoalas.simple import EventLog, Trace


DEFAULT_DELIM=" "

def convertTrace(trace: str, delimiter=DEFAULT_DELIM) -> Iterable[str]:
    if (len(trace) < 1):
        return Trace(list())
    return Trace(trace.split(delimiter))

def convert(*traces: Iterable[str], delimiter=DEFAULT_DELIM) -> EventLog:
    """
    This function converts a sequence of `"a b c"` into traces,
    and then creates an event log. Examplar uses are the following:
    \n
    `convert("a b c", "a b", "a")`\n
    `convert(*[ t for t in traces])`\n
    """
    return EventLog( map( convertTrace, traces ) )





import unittest
from logging import DEBUG

from pmkoalas.dtlog import convert
from pmkoalas.directly import DirectlyFollowPair as DFPair, FollowLanguage
from pmkoalas.directly import DIRECTLY_END,DIRECTLY_SOURCE
from pmkoalas.simple import EventLog

empty_lang = convert()

simple_lang = convert(
    "a b c",
    "a b c",
    "a d c"
)
simple_pairs = [ 
    DFPair(DIRECTLY_SOURCE, "a", 3),
    DFPair("a", "b", 2),
    DFPair("a", "d", 1),
    DFPair("b", "c", 2),
    DFPair("d", "c", 1),
    DFPair("c", DIRECTLY_END, 3),
]

many_starts = convert(
    "a e",
    "b e",
    "c e",
    "d e"
)
many_starts_pairs = [ 
    DFPair(DIRECTLY_SOURCE, "a", 1),
    DFPair(DIRECTLY_SOURCE, "b", 1),
    DFPair(DIRECTLY_SOURCE, "c", 1),
    DFPair(DIRECTLY_SOURCE, "d", 1),
    DFPair("a", "e", 1),
    DFPair("b", "e", 1),
    DFPair("c", "e", 1),
    DFPair("d", "e", 1),
    DFPair("e", DIRECTLY_END, 4),
]

many_ends = convert(
    "a b",
    "a c",
    "a d",
    "a e"
)
many_ends_pairs = [ 
    DFPair(DIRECTLY_SOURCE, "a", 4),
    DFPair("a", "b", 1),
    DFPair("a", "c", 1),
    DFPair("a", "d", 1),
    DFPair("a", "e", 1),
    DFPair("b", DIRECTLY_END, 1),
    DFPair("c", DIRECTLY_END, 1),
    DFPair("d", DIRECTLY_END, 1),
    DFPair("e", DIRECTLY_END, 1),
]

empty_traces_lang = convert(
    "a b c",
    "a b c",
    "a d c",
    "",
    "",
    ""
)
empty_traces_pairs = [ 
    DFPair(DIRECTLY_SOURCE, "a", 3),
    DFPair("a", "b", 2),
    DFPair("a", "d", 1),
    DFPair("b", "c", 2),
    DFPair("d", "c", 1),
    DFPair("c", DIRECTLY_END, 3),
]

addition_pairs = [ 
    DFPair(DIRECTLY_SOURCE, "a", 6),
    DFPair("a", "b", 4),
    DFPair("a", "d", 2),
    DFPair("b", "c", 4),
    DFPair("d", "c", 2),
    DFPair("c", DIRECTLY_END, 6),
]

class DTLogTest(unittest.TestCase):


    def check_flow_lang(self,log:EventLog,size:int,pairs, 
     debug=False):
        try:
            # check for computable and size
            flang = log.directly_follow_relations(debug=debug,
             debug_level=DEBUG) 
            self.assertEqual(len(flang), size, 
             f"expected language to be of size {size}")
            # check pairs
            self.check_directly_pairs(flang, pairs) 
        except Exception:
            self.fail()  

    def check_directly_pairs(self,flang:FollowLanguage, pairs):
        # check pairs
        computed_pairs = list(flang._relations.values())
        for pair in pairs:
            self.assertTrue(pair in pairs, f"missing pair :: {pair}")
            val = computed_pairs.index(pair)
            val = computed_pairs[val]
            self.assertTrue(val.frequency() == pair.frequency(), 
                f"{pair} :: expected frequency of {pair.frequency()}" \
                + f", but got {val.frequency()}") 

    def test_empty_lang(self):
        try:
            flang = empty_lang.directly_follow_relations()
            self.assertEqual(len(flang), 0, 
            "expected language to be empty")
        except Exception:
            self.fail()

    def test_simple_lang(self):
        self.check_flow_lang(simple_lang, 6 , simple_pairs)

    def test_many_starts(self):
        self.check_flow_lang(many_starts, len(many_starts_pairs),
         many_starts_pairs) 

    def test_many_ends(self):
        self.check_flow_lang(many_ends, len(many_ends_pairs),
         many_ends_pairs) 

    def test_empty_traces(self):
        self.check_flow_lang(empty_traces_lang, len(empty_traces_pairs),
         empty_traces_pairs)  

    def test_addition(self):
        flang = simple_lang.directly_follow_relations()
        double_flang = flang + flang
        self.check_directly_pairs(double_flang, addition_pairs) 


"""
This module has the arttributes and tags for xml documents for exporting.
"""
from typing import List
from xml.etree.ElementTree import Element 
from datetime import datetime

XES_LOG_TAG = "log"
XES_LOG_ATTRS = {
    'xes.version' : "1.0",
    'xes.features' : "nested-attributes",
    'xmlns': "http://www.xes-standard.org/"
}

# XES log sub elements
XES_EXTENSION_TAG = 'extension'
XES_EXT_CONCEPT_NAME = "concept"
XES_EXT_CONCEPT_PREFIX = "concept"
XES_EXT_CONCEPT_URI = "http://www.xes-standard.org/concept.xesext"
class XesLogExtension(Element):
    """
    Creates an xml element, for stating xes extensions used in a log.

    Parameters
    ----------
    name: `str`
    \t name of extensions
    prefix: `str`
    \t prefix used for all attributes related to extension
    uri: `str`
    \t web address for extension's xml structure
    """

    def __init__(self, name:str, prefix:str, uri:str ) -> None:
        super().__init__(
            XES_EXTENSION_TAG,
            {
                'name' : name,
                'prefix' : prefix,
                'uri' : uri
            },
            **{}
        )

XES_CLASSIFIER_TAG = "classifier"
class XesLogClassifier(Element):
    """
    Creates an xml element, for stating xes classifiers used in a log.

    Parameters
    ----------
    name: `str`
    \t name of extensions
    keys: `List[str]`
    \t the list of attributes that this classifier uses
    """

    def __init__(self, name:str, keys:List[str] ) -> None:
        super().__init__(
            XES_CLASSIFIER_TAG,
            {
                'name' : name,
                'keys' : " ".join(keys),
            },
            **{}
        )

# XES keys
XES_CONCEPT = "concept:name"
XES_TIME = "time:timestamp"

# XES sub elements
XES_STRING_TAG = 'string'
XES_INT_TAG = "int"
XES_TIME_TAG = "date"
XES_FLOAT_TAG = "float"
XES_BOOLEAN_TAG = "boolean"
class XesString(Element):
    """
    Creates an xml element, using the string representation in XES.

    Parameters
    ----------
    key: `str`
    \t The key of the attribute to be added to parent
    value: `str`
    \t The value for this key.
    """

    def __init__(self, key:str, value:str) -> None:
        super().__init__(
            XES_STRING_TAG, 
            {
                'key' : key,
                'value' : str(value)
            }, 
            **{}
        )

class XesInt(Element):
    """
    Creates an xml element, using the int representation in XES.

    Parameters
    ----------
    key: `str`
    \t The key of the attribute to be added to parent
    value: `str`
    \t The value for this key.
    """

    def __init__(self, key:str, value:int) -> None:
        super().__init__(
            XES_INT_TAG, 
            {
                'key' : key,
                'value' : str(value)
            }, 
            **{}
        )

class XesTime(Element):
    """
    Creates an xml element, using the date representation in XES.

    Parameters
    ----------
    key: `str`
    \t The key of the attribute to be added to parent
    value: `str`
    \t The value for this key.
    """

    def __init__(self, key:str, value:datetime) -> None:
        super().__init__(
            XES_TIME_TAG, 
            {
                'key' : key,
                'value' : value.isoformat()
            }, 
            **{}
        )

class XesFloat(Element):
    """
    Creates an xml element, using the float (up to eight decimals) representation in XES.

    Parameters
    ----------
    key: `str`
    \t The key of the attribute to be added to parent
    value: `str`
    \t The value for this key.
    """

    def __init__(self, key:str, value:float) -> None:
        super().__init__(
            XES_FLOAT_TAG, 
            {
                'key' : key,
                'value' : f"{value:.8f}"
            }, 
            **{}
        )

class XesBoolean(Element):
    """
    Creates an xml element, using the float (up to eight decimals) representation in XES.

    Parameters
    ----------
    key: `str`
    \t The key of the attribute to be added to parent
    value: `str`
    \t The value for this key.
    """

    def __init__(self, key:str, value:bool) -> None:
        super().__init__(
            XES_BOOLEAN_TAG, 
            {
                'key' : key,
                'value' : 'true' if value else 'false'
            }, 
            **{}
        )

# XES trace element
XES_TRACE_TAG = "trace"
class XesTrace(Element):
    """
    Creates an empty trace xml parent.

    Parameters
    ----------
    concept: `str`
    \t the concept:name value for this trace
    """

    def __init__(self, concept:str) -> None:
        super().__init__(XES_TRACE_TAG, {}, **{})
        # add concept:name
        self.append(XesString(XES_CONCEPT, concept))

# XES event element 
XES_EVENT_TAG = "event"
class XesEvent(Element):
    """
    Creates an empty event xml parent.
    """

    def __init__(self,) -> None:
        super().__init__(XES_EVENT_TAG, {}, **{})